<?php

namespace Drupal\image_deriver\Form;

use Drupal\Core\Form\FormStateInterface;

/**
 * Form for adding a Deriver.
 */
class DeriverEditForm extends DeriverFormBase {

  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    // Retrieve currently attached derivers.
    if (!$form_state->has('attached_derivers')) {
      $form_state->set('attached_derivers', $this->entity->getDerivers());
    }
    $attached_derivers = $form_state->get('attached_derivers');

    // sort the derivers by weight
    uasort($attached_derivers, function($a, $b) {
      return $a['weight'] <=> $b['weight'];
    });

    // need to wrap the whole form in a container to allow for ajax refresh
    $form['derivers_form_wrapper'] = [
      '#type' => 'container',
      '#tree' => TRUE,
      '#attributes' => ['id' => 'derivers-form-wrapper'],
    ];

    $form['derivers_form_wrapper']['responsive_image'] = [
      '#type' => 'details',
      '#title' => $this->t('Source Set Derivers'),
      '#open' => TRUE,
    ];

    // Build a table to list derivers with drag-and-drop ordering.
    $form['derivers_form_wrapper']['responsive_image']['derivers'] = [
      '#type' => 'table',
      '#tree' => TRUE,
      '#header' => [
        $this->t('Deriver'),
        ['data' => $this->t('Weight'), 'style' => 'width:60px;'],
        ['data' => $this->t('Operations'), 'style' => 'width:60px; margin-left: auto;'],
      ],
      '#attributes' => ['id' => 'derivers-table-wrapper'],
      '#tabledrag' => [
        [
          'action' => 'order',
          'relationship' => 'sibling',
          'group' => 'weight',
        ],
      ],
      '#empty' => $this->t('No derivers added yet.'),
    ];

    foreach ($attached_derivers as $delta => $deriver) {
      $form['derivers_form_wrapper']['responsive_image']['derivers'][$delta]['#attributes']['class'][] = 'draggable';
      $form['derivers_form_wrapper']['responsive_image']['derivers'][$delta]['name'] = [
          '#markup' => $deriver['id'],
      ];
      $form['derivers_form_wrapper']['responsive_image']['derivers'][$delta]['weight'] = [
        '#type' => 'weight',
        '#default_value' => $deriver['weight'],
        '#title_display' => 'invisible',
        '#attributes' => ['class' => ['weight']],
      ];
      $form['derivers_form_wrapper']['responsive_image']['derivers'][$delta]['operations'] = [
        '#type' => 'container',
        '#attributes' => [
          'style' => 'margin-left: auto;',
        ],
        'id' => [
          '#type' => 'hidden',
          '#value' => $deriver['id'],
        ],
        'delete' => [
          '#type' => 'submit',
          '#value' => $this->t('Delete'),
          '#submit' => ['::deleteDeriverSubmit'],
          // Unique button name per row.
          '#name' => 'delete_deriver_' . $deriver['id'],
          '#limit_validation_errors' => [],
          '#ajax' => [
            'callback' => '::ajaxRefreshDerivers',
            'wrapper' => 'derivers-form-wrapper',
          ],
        ],
      ];

    }

    // Inline addition of a new deriver.
    // Assume getNewDeriverOptions() returns an array of available derivers.
    $new_deriver_options = $this->getNewDeriverOptions($form_state, TRUE);
    $form['derivers_form_wrapper']['responsive_image']['new_deriver'] = [
      '#type' => 'container',
      '#tree' => TRUE, // Preserve nested structure.
      '#attributes' => ['class' => ['image-deriver-new', 'form--inline'], 'style' => ['display:flex; align-items:center;']],
    ];
    $form['derivers_form_wrapper']['responsive_image']['new_deriver']['processor'] = [
      'new' => [
        '#type' => 'select',
        '#title' => $this->t('Deriver'),
        '#title_display' => 'invisible',
        '#options' => $new_deriver_options,
        '#empty_option' => $this->t('Select a new deriver'),
      ],
      'add' => [
        '#type' => 'submit',
        '#value' => $this->t('Add'),
        '#validate' => ['::validateNewDeriver'],
        '#submit' => ['::submitNewDeriver'],
        '#limit_validation_errors' => [['derivers_form_wrapper', 'responsive_image', 'new_deriver', 'processor', 'new']],
        '#ajax' => [
          'callback' => '::ajaxRefreshDerivers',
          'wrapper' => 'derivers-form-wrapper',
        ],
      ],
    ];

    $form['derivers_form_wrapper']['single_image'] = [
      '#type' => 'details',
      '#title' => $this->t('Image Tag Deriver'),
      '#tree' => TRUE, // Preserve nested structure.
      '#open' => TRUE,
    ];

    // get the selected single image deriver.

    // add a select for the single image deriver
    $form['derivers_form_wrapper']['single_image']['deriver'] = [
      '#type' => 'select',
      '#title' => $this->t('Image Deriver'),
      '#options' => $this->getNewDeriverOptions($form_state),
      '#empty_option' => $this->t('Select a deriver'),
      '#default_value' => $this->entity->getSingleImageDeriver(),
    ];

    return $form;
  }


  protected function getNewDeriverOptions(FormStateInterface $form_state, $remove_attached = FALSE) {
    $plugin_manager = \Drupal::service('image_deriver.plugin_manager');
    $definitions = $plugin_manager->getDefinitions();

    if ($remove_attached) {
      // remove any derivers that are already attached
      foreach ($this->entity->getDerivers() as $deriver) {
        unset($definitions[$deriver['id']]);
      }

      // also remove any derivers that are added in the form_state
      if ($form_state->has('attached_derivers')) {
        foreach ($form_state->get('attached_derivers') as $deriver) {
          unset($definitions[$deriver['id']]);
        }
      }
    }

    $options = [];
    foreach ($definitions as $plugin_id => $definition) {
      $options[$plugin_id] = isset($definition['label']) ? $definition['label'] : $plugin_id;
    }

    return $options;
  }

  public function validateNewDeriver(array &$form, FormStateInterface $form_state) {
    $deriver = $form_state->getValue(['derivers_form_wrapper', 'responsive_image', 'new_deriver', 'processor', 'new']);
    if (empty($deriver)) {
      $form_state->setErrorByName('new', $this->t('You must select a deriver to add.'));
    }

    // also ensure that the deriver is not already attached
    if ($this->entity->getDeriver($deriver)) {
      $form_state->setErrorByName('new', $this->t('Deriver %name is already attached.', ['%name' => $deriver]));
    }
  }

  public function submitNewDeriver(array &$form, FormStateInterface $form_state) {
    $attached_derivers = $form_state->get('attached_derivers');
    $deriver = $form_state->getValue(['derivers_form_wrapper', 'responsive_image', 'new_deriver', 'processor', 'new']);
    // Add the new deriver to the temporary state.
    $attached_derivers[$deriver] = [
      'id' => $deriver,
      'weight' => 0,
      'data' => []
    ];
    $form_state->set('attached_derivers', $attached_derivers);
    $this->messenger()->addStatus($this->t('Deriver %name has been added.', ['%name' => $deriver]));
    $form_state->setRebuild(TRUE);
  }

  public function ajaxRefreshDerivers(array &$form, FormStateInterface $form_state) {
    return $form['derivers_form_wrapper'];
  }

  public function deleteDeriverSubmit(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $deriver = substr($trigger['#name'], strlen('delete_deriver_'));
    $attached_derivers = $form_state->get('attached_derivers');
    unset($attached_derivers[$deriver]);
    $form_state->set('attached_derivers', $attached_derivers);
    $this->messenger()->addStatus($this->t('Deriver %name has been removed.', ['%name' => $deriver]));
    $form_state->setRebuild(TRUE);
  }

  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);

    // Retrieve the current temporary state of derivers.
    $attached_derivers = $form_state->get('attached_derivers');
    // Get the submitted values for each deriver row (which include the updated weight).
    $submitted_derivers = $form_state->getValue(['derivers_form_wrapper', 'responsive_image', 'derivers']);

    // Update the temporary state with the new weight values.
    foreach ($submitted_derivers as $delta => $values) {
      if (isset($attached_derivers[$delta])) {
        $attached_derivers[$delta]['weight'] = $values['weight'];
      }
    }

    // Get the selected single image deriver.
    $single_image_deriver = $form_state->getValue(['derivers_form_wrapper', 'single_image', 'deriver']);
    $this->entity->setSingleImageDeriver($single_image_deriver);

    // Update the entity with the modified derivers and save.
    $this->entity->setDerivers($attached_derivers);
    $this->entity->save();
    $this->messenger()->addStatus($this->t('Deriver %name has been saved.', ['%name' => $this->entity->label()]));
  }

}
