<?php

declare(strict_types=1);

namespace Drupal\image_deriver;

use Drupal\Component\Utility\SortArray;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\ImageToolkit\ImageToolkitManager;
use Drupal\Core\Template\AttributeString;
use Drupal\image\Entity\ImageStyle;
use Drupal\image_deriver\Entity\ImageDeriver;
use Drupal\responsive_image\Entity\ResponsiveImageStyle;

/**
 * @todo Add class description.
 */
final class ImageDeriverPreprocessor {

  /**
   * Constructs an ImageDeriverPreprocessor object.
   */
  public function __construct(
    private readonly FileUrlGeneratorInterface $fileUrlGenerator,
    private readonly ImageDeriverPluginManager $imageDeriverPluginManager,
    private readonly ImageToolkitManager $imageToolkitManager,
  ) {}

  protected function getOriginalImageFormat($uri) {
    $format = pathinfo($uri, PATHINFO_EXTENSION);
    // remove the query string
    if (strpos($format, '?') !== false) {
      $format = substr($format, 0, strpos($format, '?'));
    }

    $format = $this->imageToolkitManager->getDefaultToolkit()->extensionToImageType($format);

    $format = image_type_to_extension($format, FALSE);

    return $format;
  }

  public function preprocessImage(&$variables) {
    $image_deriver_id = $variables['attributes']['image_deriver'] ?? FALSE;
    if ($image_deriver_id) {
      $uri = $variables['attributes']['original_image_uri'] ?? $variables['uri'];
      $image_deriver = ImageDeriver::getDeriverById($image_deriver_id);
      $image_styles = [];
      if (!$image_deriver) {
        return;
      }

      $image_deriver_plugin_id = $image_deriver->getSingleImageDeriver();
      $image_deriver_plugin = $this->imageDeriverPluginManager->getDefinition($image_deriver_plugin_id);
      $image_deriver_id = $image_deriver_plugin['id'] ?? FALSE;

      if ($image_deriver_id === 'original') {
        // short circuit if the deriver is original.
        return;
      }

      $image_deriver_format = $image_deriver_plugin['format'] ?? FALSE;
      $image_derivers = [$image_deriver_id => $image_deriver_plugin];
      // get the plugin definition
      $image_style = $variables['style_name'] ?? FALSE;
      if ($image_style) {
        $image_styles = [ImageStyle::load($image_style)];
      }

      if (!$image_style) {
        $responsive_image_style = $variables['attributes']['responsive_image_style_id'] ?? FALSE;
        if ($responsive_image_style) {
          $responsive_image_style = ResponsiveImageStyle::load($responsive_image_style);
          $image_styles = ImageStyle::loadMultiple($responsive_image_style->getImageStyleIds() ?? []); // @todo Check if this is correct.
        }
      }

      $original_uri = $variables['attributes']['original_image_url'] ?? FALSE;
      if ($original_uri) {
        $uri = $original_uri;
      }

      if ($image_styles && $image_deriver_format) {
        $image_style_map = [];
        foreach ($image_styles as $image_style) {
          $this->buildImageStyleMap($image_style_map, $uri, $image_style, $image_derivers);
        }

        if ($image_style_map) {
          $variables['attributes']['src'] = strtr($variables['attributes']['src'], $image_style_map[$image_deriver_format]);

          /** @var \Drupal\Core\Template\AttributeString $srcset */
          $srcset = $variables['attributes']['srcset'] ?? FALSE;

          if ($srcset instanceof AttributeString) {
            $variables['attributes']['srcset'] = strtr($srcset->value(), $image_style_map[$image_deriver_format]);
          }
        }

      }

      unset($variables['attributes']['image_deriver']);
      unset($variables['attributes']['responsive_image_style_id']);
      unset($variables['attributes']['original_image_uri']);
    }
  }

  /**
   * Preprocesses the responsive image tag.
   *
   * @param array $variables
   *  The variables to preprocess.
   *
   * @return void
   */
  public function preprocessResponsive(&$variables): void {
    // @todo Determine if we should get rid of service if replaced by OOP hook.
    // @todo Test for non picture tags
    $image_deriver_id = $variables['attributes']['image_deriver'] ?? FALSE;
    if ($image_deriver_id) {
      $image_deriver = ImageDeriver::getDeriverById($image_deriver_id);

      if (!$image_deriver) {
        return;
      }

      $image_derivers = $image_deriver->getDerivers();
      uasort($image_derivers, [SortArray::class, 'sortByWeightElement']);
      $image_style_map = [];
      $entity = ResponsiveImageStyle::load($variables['responsive_image_style_id']);

      // add to the img_element attributes the responsive_image_style_id so we can use it for the single image logic.
      $variables['img_element']['#attributes']['responsive_image_style_id'] = $variables['responsive_image_style_id'];

      // add original image context
      $variables['img_element']['#attributes']['original_image_url'] = $variables['uri'];

      foreach (ImageStyle::loadMultiple($entity->getImageStyleIds()) as $image_style) {
        $this->buildImageStyleMap($image_style_map, $variables['uri'], $image_style, $image_derivers);
      }
      if ($image_style_map) {
        $sources = [];
        foreach ($variables['sources'] as $source) {
          $offset_key = NULL;
          // Check for instances where srcset and data-srcset used interchangeably.
          if ($source->offsetExists('srcset')) {
            $offset_key = 'srcset';
          }
          elseif ($source->offsetExists('data-srcset')) {
            $offset_key = 'data-srcset';
          }
          // If either key is found, continue.
          if ($offset_key) {
            $source_srcset = $source->offsetGet($offset_key)->value();
              foreach($image_style_map as $format => $derivatives) {
                if ($format === 'original') {
                  $sources[] = $source;
                  continue;
                }
                $cloned_srcset = strtr($source_srcset, $derivatives);
                $source_derivative = clone $source;
                $source_derivative->offsetSet($offset_key, $cloned_srcset);
                $source_derivative->offsetSet('type', 'image/' . $format);
                $sources[] = $source_derivative;
            }
          }
          else {
            $sources[] = $source;
          }
        }
        $variables['sources'] = $sources;
      }

      // This is needed to render the sources.
      $variables['output_image_tag'] = FALSE;

      // @todo Convert fallback image to single image style(if defined).
      // @todo Do we need any cache keys?
    }
  }

  /**
   * @param array $image_style_map
   * @param string $uri
   * @param ImageStyle $image_style
   * @param array $image_derivers
   *
   * @return array
   */
  protected function buildImageStyleMap(array &$image_style_map, string $uri, ImageStyle $image_style, array $image_derivers) {
    // clean the uri of query parameters
    $scheme = parse_url($uri, PHP_URL_SCHEME);
    if ($scheme !== 'http' && $scheme !== 'https') {
      $image_uri = $image_style->buildUri($uri);
      $image_url = $this->fileUrlGenerator->generateAbsoluteString($image_uri);
    } else {
      $image_url = $uri;
    }
    $image_relative_path = $this->fileUrlGenerator->transformRelative($image_url);

    // check if the relative path already has a query string param itok. if not we can add it.
    if (strpos($image_relative_path, 'itok') === FALSE) {
      $image_relative_path .= (strpos($image_relative_path, '?') !== FALSE ? '&' : '?') . 'itok=' . $image_style->getPathToken($uri);
    }

    $original_image_format = $this->getOriginalImageFormat($uri);


    foreach ($image_derivers as $key => $deriver) {
      $deriver_id = $deriver['id'];
      $plugin_definition = $this->imageDeriverPluginManager->getDefinition($deriver_id);
      $format = $plugin_definition['format'] ?? FALSE;

      // skip if format matches the original image format.
      if (!$format || $format === $original_image_format) {
        continue;
      }

      // sometimes the uri can be absolute and contain the itok query string.
      $uri = strtok($uri, '?');
      if ($scheme !== 'http' && $scheme !== 'https') {
        $image_derivative_uri = $image_style->buildUri($uri . '.' . $format);
        $image_derivative_url = $this->fileUrlGenerator->generateAbsoluteString($image_derivative_uri);
      } else {
        $image_derivative_url = $uri . '.' . $format;
      }

      $image_derivative_relative_path = $this->fileUrlGenerator->transformRelative($image_derivative_url);
      $image_derivative_relative_path .= (strpos($image_derivative_relative_path, '?') !== FALSE ? '&' : '?') . 'itok=' . $image_style->getPathToken($uri . '.' . $format);
      $image_style_map[$format][$image_relative_path] = $image_derivative_relative_path;
    }

    return $image_style_map;
  }
}
